"""A simple logging library, designed for deep learning.

Trainlog's aim is to make it easy to create a single file containing all
key info about a training run. We prioritize convenience and ease of use
over efficiency.

**Example**

This is a minimal example of training and analysis using trainlog.
For a more complete example, see `trainlog.examples.pytorch`.

First some aliased imports. The second is optional (`L.ops.*` would also
work).

>>> import trainlog as L
>>> import trainlog.ops as O

In training, we open a log, add some custom info to the header event,
`name="example"`, and emit some training step events of kind "step".

>>> with L.logger.open("log.jsonl", name="example") as logger:
...     logger.add("step", loss=10)
...     logger.add("step", loss=5)
...     logger.add("step", loss=2.5)

In post-hoc analysis we open the log file, which has been gzipped at the
end of training, hence the `.gz` extension.

>>> log = L.logs.open("log.jsonl.gz")

We transform the log by copying `"name"` from the header, and computing a
step counter. Then we select just the `"step"` events and convert them to a
pandas DataFrame for further analysis.

>>> log = log.apply(O.header("name"), O.count("step"))
>>>
>>> df = log["step"].to_pandas()
>>> df[["step", "loss"]]
   step  loss
0     0  10.0
1     1   5.0
2     2   2.5
"""

from . import io, logger, logs, ops  # noqa: F401
from ._version import __version__  # noqa: F401

__pdoc__ = dict(tests=False)



"""Defines operations that transform log event sequences.

Includes a generic abstraction BaseOperation, for any transformation
of `Iterable[Event] -> Iterable[Event]`.

Most operations are derived from `Operation`, which defines an
operation that starts with an initial value and reduces over a log,
updating events as it goes.

For example, this operation can count the number of events that have
{"kind": "step"}:

    op = ops.Sum(ops.kind("step"), "step")
    events_out = op(events_in)

There is also a "functional" API that provides some convenient defaults
and conversions. Since `kind("...")` is such a common case, there is an
automatic conversion from string. Also, `kind()` provides a default name
for the output key, so the above `op` is equivalent to:

    op = ops.sum("step")
"""

import abc
import builtins
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    TypeVar,
    Union,
)

# pylint:disable=redefined-builtin

T = TypeVar("T")
Event = Dict[str, Any]

Mapping = Callable[[Event], Any]
Predicate = Callable[[Event], bool]
AutoPredicate = Union[str, Predicate]
ScalarFn = Callable[[Event], Union[bool, float]]
AutoScalarFn = Union[str, ScalarFn]

Reduction = Callable[[Iterable[Event]], Any]
BaseOperation = Callable[[Iterator[Event]], Iterator[Event]]

# Operations


class Operation(abc.ABC):
    """Base class for log denormalization operations.

    An operation is a small program that is run over a sequence of events from a log:

        value = operation.initial()
        for event in events:
            operation.update(event, value)
            value = operation.accumulate(event, value)
    """

    # pylint:disable=no-self-use

    def initial(self) -> Any:
        """[Optional] An initial value for the reduction."""
        return None

    def accumulate(
        self, event: Event, value: Any  # pylint:disable=unused-argument
    ) -> Any:
        """[Optional] A reducing function to generate a new value."""
        return None

    @abc.abstractmethod
    def update(self, event: Event, value: Any) -> None:
        """Applies the operation to update the given event.

        This operation is run before the event is `accumulate`d into `value`.

        The default implementation of `__call__` makes a shallow copy of `event`
        externally, so that it's safe to add or remove toplevel keys, however the
        `update()` implementation is responsible for deep copies, if required.

        Note that this may not be called for every event that accumulate() is called on
        (see `When`).
        """
        raise NotImplementedError

    def __call__(self, events: Iterator[Event]) -> Iterator[Event]:
        """Run the operation over an event stream, yielding a new stream."""
        # pylint:disable=assignment-from-none
        value = self.initial()
        for event in events:
            event = event.copy()
            self.update(event, value)
            yield event
            value = self.accumulate(event, value)


class Map(Operation):
    """Apply an elementwise mapping function."""

    def __init__(self, mapping: Mapping, name: str):
        self.mapping = mapping
        self.name = name

    def update(self, event: Event, value: None) -> None:
        event[self.name] = self.mapping(event)


class Copy(Operation):
    """Copy a value from a previous event."""

    def __init__(self, predicate: Predicate, src: str, dest: str):
        self.predicate = predicate
        self.src = src
        self.dest = dest

    def accumulate(self, event: Event, value: Any) -> Any:
        return event[self.src] if self.predicate(event) else value

    def update(self, event: Event, value: Any) -> None:
        if not self.predicate(event):
            event[self.dest] = value


class Sum(Operation):
    """Compute a scalar sum of previous events.

    E.g. counting events of a specific kind.
    """

    def __init__(self, scalarfn: ScalarFn, name: str):
        self.scalarfn = scalarfn
        self.name = name

    def initial(self) -> float:
        return 0

    def accumulate(self, event: Event, value: float) -> float:
        return value + self.scalarfn(event)

    def update(self, event: Event, value: float) -> None:
        event[self.name] = value


class Window(Operation):
    """Aggregate over a sliding window of preceding events matching a predicate."""

    def __init__(
        self, predicate: Predicate, size: Optional[int], reduction: Reduction, name: str
    ):
        self.predicate = predicate
        self.size = size
        self.reduction = reduction
        self.name = name

    def initial(self) -> List[Event]:
        return []

    def accumulate(self, event: Event, value: List[Event]) -> List[Event]:
        if self.predicate(event):
            value.append(event)
            if self.size is not None and len(value) > self.size:
                value = value[-self.size :]
        return value

    def update(self, event: Event, value: List[Event]) -> None:
        event[self.name] = self.reduction(value)


class Group(Operation):
    """Create a single operation that applies multiple operations in series."""

    def __init__(self, operations: Iterable[Operation]):
        self.operations = operations

    def initial(self) -> List[Any]:
        return [op.initial() for op in self.operations]

    def accumulate(self, event: Event, value: List[Any]) -> List[Any]:
        return [
            op.accumulate(event, opval) for op, opval in zip(self.operations, value)
        ]

    def update(self, event: Event, value: List[Any]) -> None:
        for op, opval in zip(self.operations, value):
            op.update(event, opval)


class When(Operation):
    """[Wrapper] Conditionally run update() for an operation.

    E.g. with `kind("step")` to only update step events
    """

    def __init__(self, predicate: Predicate, body: Operation):
        self.predicate = predicate
        self.body = body

    def initial(self) -> Any:
        return self.body.initial()

    def accumulate(self, event: Event, value: Any) -> Any:
        return self.body.accumulate(event, value)

    def update(self, event: Event, value: Any) -> None:
        if self.predicate(event):
            self.body.update(event, value)


class Duck(Operation):
    """Swallow key errors to make an operation "duck typed" on the event.

    Note: if the wrapped event raises a `KeyError`, it must leave `event`
    unchanged.
    """

    def __init__(self, body: Operation):
        self.body = body

    def initial(self) -> Any:
        return self.body.initial()

    def accumulate(self, event: Event, value: Any) -> Any:
        try:
            return self.body.accumulate(event, value)
        except KeyError:
            return value

    def update(self, event: Event, value: Any) -> None:
        try:
            self.body.update(event, value)
        except KeyError:
            pass


def _auto_name(
    name: Optional[str], fn: Callable[[Any], Any], prefix: str, context: str
) -> str:
    if name is not None:
        return name
    if fn.__name__ in {None, "<lambda>"}:
        raise ValueError(
            f"If `{context}()` using a lambda, you must provide a `name=...`"
        )
    return prefix + fn.__name__


def filter(predicate: AutoPredicate) -> BaseOperation:
    """An operation to filter an event stream by predicate or kind.

    For example:

        filter("valid")
        filter(lambda event: event["step"] >= 100)
    """
    pred = to_predicate(predicate)
    return lambda events: builtins.filter(pred, events)


def map(mapping: Mapping, name: Optional[str] = None) -> Operation:
    """An operation to apply a "pointwise" mapping function to events.

    The default `name` is `mapping.__name__`, unless None or "<lambda>".

    For example:

        map(lambda event: 1 - event["error_rate"], "accuracy")
    """
    return Map(mapping, _auto_name(name, mapping, prefix="", context="map"))


def copy(predicate: AutoPredicate, src: str, dest: Optional[str] = None) -> Operation:
    """An operation to copy a field from a previous event.

    The default `dest` is `src`.

    For example:

        copy("valid", "loss", "last_valid_loss")

    Equivalent to:

        value = None
        for event in events:
            if predicate(event):
                value = event[src]
            else:
                event[dest] = value
    """
    return Copy(to_predicate(predicate), src, dest or src)


def header(src: str, dest: Optional[str] = None) -> Operation:
    """An operation to copy a field from the header.

    Equivalent to `copy("header", src, dest)`.

    For example:

        header("id")
        header("learning_rate", "lr")
    """
    return Copy(kind("header"), src, dest or src)


def sum(scalarfn: AutoScalarFn, name: Optional[str] = None) -> Operation:
    """An operation to sum values from previous events.

    The default `name` is `sum_{scalarfn.__name__}`, unless None or "<lambda>".

    For example:

        sum(ops.get("examples"))
        sum(lambda event: event["examples"], "sum_examples")
    """
    scalarfn = to_predicate(scalarfn)
    return Sum(scalarfn, _auto_name(name, scalarfn, prefix="sum_", context="sum"))


def count(predicate: AutoPredicate, name: Optional[str] = None) -> Operation:
    """An operation to count occurrences of previous events.

    The default `name` is `predicate.__name__`, unless None or "<lambda>".

    For example:

        count("step")
        count(lambda event: event["loss"] > 10, "n_large_losses")
    """
    predicate = to_predicate(predicate)
    return Sum(predicate, _auto_name(name, predicate, prefix="", context="count"))


def window(
    predicate: AutoPredicate,
    size: Optional[int],
    reduction: Reduction,
    name: Optional[str] = None,
) -> Operation:
    """An operation to compute a statistic over a window of previous events.

    The default `name` is `reduction.__name__`, unless None or "<lambda>".

    For example:

        window("step", 10, ops.reduce_mean("loss"), "train_loss")
        window("step", 10,
            lambda events: np.mean([e["loss"] for e in events]),
            "train_loss")

    Equivalent to:

        previous = []
        for event in events:
            event[name] = reduction(previous)
            if predicate(event):
                previous = (previous + [event])[-size:]
    """
    return Window(
        to_predicate(predicate),
        size,
        reduction,
        _auto_name(name, reduction, prefix="", context="window"),
    )


def when(predicate: AutoPredicate, body: Operation) -> Operation:
    """An operation wrapper that only updates events when `predicate(event)`.

    For example:

        when("valid", ops.map(lambda event: 1 - event["error_rate"], "accuracy"))
    """
    return When(to_predicate(predicate), body)


def group(*operations: Operation) -> Operation:
    """An operation wrapper to apply multiple operations in a single pass.

    All `op.update()` are run first, followed by `op.accumulate()`.

    For example:

        group(ops.header("id"), ops.count("step"))
    """
    return operations[0] if len(operations) == 1 else Group(operations)


def duck(operation: Operation) -> Operation:
    """An operation wrapper to catch and ignore any KeyError from the operation.

    This can be used to "duck type" an operation & only apply it to events with
    the expected keys.

    For example:

        duck(ops.map(lambda event: 1 - event["error_rate"], "accuracy"))
    """
    return Duck(operation)


# Predicate/Scalarfn


def kind(value: str) -> Predicate:
    """A predicate to match events with `{"kind": value}`."""

    def predicate(event: Event) -> bool:
        return event.get("kind") == value

    predicate.__name__ = value
    return predicate


def to_predicate(
    kind_or_fn: Union[str, Callable[[Event], T]]
) -> Callable[[Event], Union[bool, T]]:
    """Automatic conversion from string to predicate `kind(kind_or_fn)`."""
    if isinstance(kind_or_fn, str):
        return kind(kind_or_fn)
    return kind_or_fn


def get(key: str, required: bool = False) -> ScalarFn:
    """A scalarfn for use with `sum` that gets a key from the event.

    If `required`, raise a KeyError if the key is missing (maybe useful with `duck`),
    otherwise returns `None`.

    Functionally equivalent to `operators.itemgetter(key)` when required is True,
    or `lambda e: e.get(key)` if required is False.

    Unlike lambdas or `operators`, this sets `__name__` so that there is a sensible
    default name for the generated event.

    For example:

        ops.sum(get("examples"))  # sums "examples" => "sum_examples"
    """

    def scalarfn(event: Event) -> Any:
        return event[key] if required else event.get(key)

    scalarfn.__name__ = key
    return scalarfn


# Reductions


def reduce_mean(key: str) -> Callable[[Iterable[Event]], Optional[float]]:
    """A reduction for use with `window`, that computes the mean of a scalar key.

    If passed an empty list, returns `None`.
    """

    def reduce(events: Iterable[Event]) -> Optional[float]:
        total = 0
        n = 0
        for event in events:
            total += event[key]
            n += 1
        return total / n if n else None

    reduce.__name__ = f"mean_{key}"
    return reduce



"""Core IO abstractions - reading & writing JSON Lines (https://jsonlines.org/).

This module provides `JsonLinesIO`, a simplified stream reader/writer for JSON lines,
which also adds support for serializing/deserializing numpy arrays.
"""

from __future__ import annotations

import functools as ft
import gzip as gzip_
import json
import os
import typing
from types import TracebackType
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    Optional,
    TextIO,
    Tuple,
    Type,
    TypeVar,
)

T = TypeVar("T")

NUMPY_DICT_KEY = "__numpy_dict"


def numpy_to_dict(array: Any) -> Dict[str, Any]:
    """Convert a numpy array to a JSON-able dictionary, `numpy_from_dict` restores."""
    return {
        NUMPY_DICT_KEY: 0,
        "shape": list(array.shape),
        "dtype": array.dtype.base.name,
        "items": array.flatten().tolist(),
    }


def numpy_from_dict(dict_: Dict[str, Any]) -> Any:
    """Convert a JSON-able dictionary from `numpy_to_dict` back to a numpy array.

    This test identifies valid `dict`:

        if trainlog.io.NUMPY_DICT_KEY in dict_:
            array = numpy_from_dict(dict_)
    """
    import numpy as np  # type: ignore  # pylint: disable=import-outside-toplevel

    assert dict_[NUMPY_DICT_KEY] == 0
    shape = tuple(dict_["shape"])
    dtype = np.dtype(dict_["dtype"])
    items = dict_["items"]
    return np.array(items, dtype=dtype).reshape(shape)


class JSONEncoderWithNumpy(json.JSONEncoder):
    """Add numpy array support using `numpy_to_dict`."""

    def default(self, o: Any) -> Any:
        if type(o).__name__ == "ndarray":
            return numpy_to_dict(o)
        return super().default(o)


ObjectHook = Callable[[Dict[str, Any]], Any]
ObjectPairsHook = Callable[[Iterable[Tuple[str, Any]]], Any]


class JSONDecoderWithNumpy(json.JSONDecoder):
    """Add numpy array support using `numpy_from_dict`."""

    @staticmethod
    def create_object_hook(
        next_hook: Optional[ObjectHook], dict_: Dict[str, Any]
    ) -> Any:
        """Create a chained object_hook, which tries `numpy_from_dict` first."""
        if NUMPY_DICT_KEY in dict_:
            return numpy_from_dict(dict_)
        if next_hook is not None:
            return next_hook(dict_)
        return dict_

    @staticmethod
    def create_object_pairs_hook(
        next_hook: ObjectPairsHook, obj_pairs: Iterable[Tuple[str, Any]]
    ) -> Any:
        """Create a chained object_pairs_hook, which tries `numpy_from_dict` first."""
        dict_ = dict(obj_pairs)
        if NUMPY_DICT_KEY in dict_:
            return numpy_from_dict(dict_)
        return next_hook(obj_pairs)

    def __init__(
        self,
        *,
        object_hook: Optional[ObjectHook] = None,
        object_pairs_hook: Optional[ObjectPairsHook] = None,
        **args: Any
    ):
        object_pairs_hook = (
            ft.partial(self.create_object_pairs_hook, object_pairs_hook)
            if object_pairs_hook is not None
            else None
        )
        super().__init__(
            object_hook=ft.partial(self.create_object_hook, object_hook),
            object_pairs_hook=object_pairs_hook,
            **args
        )


class JsonLinesIO(Generic[T]):
    """Reader/writer for JSON Lines files.

    See https://jsonlines.org/.

    Similar to `TextIO`, but writes "JSON-able" objects rather than strings.
    """

    stream: TextIO
    dump_args: Dict[str, Any]
    load_args: Dict[str, Any]

    def __init__(
        self,
        stream: TextIO,
        dump_args: Optional[Dict[str, Any]] = None,
        load_args: Optional[Dict[str, Any]] = None,
    ):
        self.stream = stream
        self.dump_args = dump_args.copy() if dump_args else {}
        self.dump_args.setdefault("separators", (",", ":"))
        self.dump_args.setdefault("cls", JSONEncoderWithNumpy)
        self.load_args = load_args.copy() if load_args else {}
        self.load_args.setdefault("cls", JSONDecoderWithNumpy)

    def __enter__(self) -> JsonLinesIO[T]:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()

    def __iter__(self) -> Iterator[T]:
        return self.objects()

    def close(self) -> None:
        """Close the underlying text stream."""
        self.stream.close()

    def flush(self) -> None:
        """Flush the underlying text stream."""
        self.stream.flush()

    def write(self, obj: T) -> None:
        """Write an object to the file, as a JSON entry on a single line."""
        json.dump(obj, self.stream, **self.dump_args)
        self.stream.write("\n")

    def read(self) -> T:
        """Read a single object from the file.

        Throws EOFError if there are no more JSON objects in the file.
        """
        line = self.stream.readline()
        if not line:
            raise EOFError(
                "Attempting to read JSON data past the end of stream", self.stream
            )
        return typing.cast(T, json.loads(line, **self.load_args))

    def objects(self) -> Iterator[T]:
        """An iterator over objects in the file."""
        try:
            while True:
                yield self.read()
        except EOFError:
            pass


def open_maybe_gzip(
    path: str,
    mode: str = "r",
    gzip: Optional[bool] = None,  # pylint: disable=redefined-outer-name
) -> TextIO:
    """Open a file, but use gzip.open if appropriate.

    gzip -- Treat the file as GZIP? If `None`, autodetect based on path extension.
    """
    if gzip or (gzip is None and os.path.splitext(path)[-1] in (".gz", ".gzip")):
        # Mode should default to text, for consistency with `open()`
        gzip_mode = mode if "b" in mode or "t" in mode else mode + "t"
        return typing.cast(TextIO, gzip_.open(path, gzip_mode))
    return typing.cast(TextIO, open(path, mode))


def read_jsonlines(
    path: str, load_args: Optional[Dict[str, Any]] = None
) -> Iterator[T]:
    """Read JSON Lines from a local filesystem path."""
    with JsonLinesIO[T](open_maybe_gzip(path), load_args=load_args) as reader:
        yield from reader


def write_jsonlines(
    path: str, objects: Iterable[T], dump_args: Optional[Dict[str, Any]] = None
) -> None:
    """Write JSON Lines to a local filesystem path."""
    with JsonLinesIO[T](open_maybe_gzip(path, "w"), dump_args=dump_args) as writer:
        for obj in objects:
            writer.write(obj)


def gzip(
    path: str, extension: str = ".gz", delete: bool = True, chunk_size: int = 1024
) -> None:
    """Gzip a local file (by default deleting the original afterwards)."""
    assert extension, "cannot write the gzip to the file being read"
    with open(path, "rb") as srcf, gzip_.open(str(path) + extension, "wb") as destf:
        buffer = bytearray(chunk_size)
        while True:
            count = srcf.readinto(buffer)  # type: ignore
            if not count:
                break
            destf.write(buffer[:count])
    if delete:
        os.remove(path)



"""Friendly APIs for preprocessing log files.

Note that it's simple to handle log files directly, using `io.read_jsonlines`.
This API adds a few tools for dealing with heterogenous event streams, before
handing them off to a tabular data processing library such as pandas.

For example, we might write the following to generate data that's easy to work
with in pandas:

    import trainlog.ops as O

    logs = trainlog.logs.glob("results/*.jsonl.gz")

    logs = logs.apply(
        O.header("learning_rate"),
        O.count_if("step"),
        O.when("valid", O.window("step", 100, O.reduce_mean("loss"), "train_loss")),
    )

    df = logs["valid"].to_pandas()
"""

from __future__ import annotations

import datetime
import glob as pyglob
import os
import typing
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

from . import io, ops
from .ops import Event


class JsonLinesFile:
    """An event stream that is lazily read from a JSONlines file.

    Note that if the first event in the file has {"kind": "header"}, this class
    automatically adds a key {"metadata": {"path": ..., "created": ...,
    "modified": ...}}.
    """

    def __init__(self, path: str, load_args: Optional[Dict[str, Any]] = None):
        self.path = path
        self.load_args = load_args

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.path!r})"

    def __iter__(self) -> Iterator[Event]:
        iterator: Iterator[Event] = iter(io.read_jsonlines(self.path, self.load_args))
        try:
            first_event = next(iterator)
            if first_event.get("kind") == "header":
                first_event["metadata"] = self.metadata
            yield first_event
            yield from iterator
        except StopIteration:
            pass

    @property
    def metadata(self) -> Dict[str, Any]:
        """A dictionary of metadata about the file loaded."""
        return dict(
            path=self.path,
            created=datetime.datetime.fromtimestamp(
                os.path.getctime(self.path)
            ).isoformat(),
            modified=datetime.datetime.fromtimestamp(
                os.path.getmtime(self.path)
            ).isoformat(),
        )


def _events_repr(events: Iterable[Event]) -> str:
    """A string summary for tuples or lists that doesn't print all the contents."""
    if isinstance(events, (tuple, list)):
        return f"[{len(events)}]"
    return repr(events)


class Transform:
    """An event stream produced by transforming another stream.

    If we compare two alternatives:

        operation(events)
        Transform(events, operation)

    The main difference is that the second can be iterated multiple times (as long
    as `events` can.)
    """

    def __init__(self, events: Iterable[Event], operation: ops.BaseOperation):
        self.events = events
        self.operation = operation

    def __repr__(self) -> str:
        return f"{type(self).__name__}({_events_repr(self.events)}, {self.operation!r})"

    def __iter__(self) -> Iterator[Event]:
        return self.operation(iter(self.events))


def get_header(events: Iterable[Event]) -> Optional[Event]:
    """Extract the header from the event stream.

    The header must be the first event, and have {"kind": "header"}.
    """
    first = next(iter(events), None)
    if first is None:
        return None
    if first.get("kind") != "header":
        return None
    return first


def list_to_array(column: Sequence[Any]) -> Any:
    """Convert a list to a numpy array, with a slight tweak to np.asarray.

    The differences between this & np.asarray is the handling of None, and
    the ability to output multidimensional arrays.

        list_to_array([1, None]) == np.array([1.0, np.nan])

        list_to_array([[1], [2]]) \
            == np.array([np.array([1]), np.array([2])], dtype=np.object)
    """
    import numpy as np  # type: ignore  # pylint: disable=import-outside-toplevel

    types = {type(value) for value in column}

    if {np.ndarray, tuple, list} & types:
        # Deep structure - this could be a pain, so handle it separately.
        # We don't want to generate ND structures anyway, as pandas won't
        # handle these
        objarray = np.full(len(column), None)
        objarray[:] = column
        return objarray

    # Use dtype=np.float to convert None to np.nan as long as all types are
    # numeric
    dtype = None
    if type(None) in types and all(
        t is type(None) or np.dtype(t).kind in "biuf" for t in types  # noqa: E721
    ):
        dtype = np.float

    return np.asarray(column, dtype=dtype)


class Columns:
    """A set of named columns, stored as arrays (requires numpy)."""

    def __init__(
        self, columns: Dict[str, Any], column_order: Optional[Tuple[str, ...]]
    ):
        self.columns = columns
        self.column_order = column_order
        for name, column in columns.items():
            attrname = f"{name}_" if hasattr(self, name) else name
            setattr(self, attrname, column)

    def __repr__(self) -> str:
        if self.column_order is not None:
            return f"{type(self).__name__}{self.column_order!r}"
        return f"{type(self).__name__}{tuple(self.columns)!r}"

    def __getitem__(self, key_or_index: Union[str, int]) -> Any:
        if isinstance(key_or_index, str):
            return self.columns[key_or_index]
        if self.column_order is None:
            raise ValueError(
                "Cannot use Columns as an indexable sequence unless it has"
                " a defined column ordering"
            )
        return self.columns[self.column_order[key_or_index]]

    def __len__(self) -> int:
        return len(self.columns)

    def to_pandas(self) -> Any:
        """Convert to a pandas DataFrame."""
        import pandas  # type: ignore  # pylint: disable=import-outside-toplevel

        df = pandas.DataFrame.from_dict(self.columns)
        if self.column_order is not None:
            df = df[list(self.column_order)]
        return df

    @classmethod
    def from_events(
        cls, events: Iterable[Event], columns: Optional[Sequence[str]] = None
    ) -> Columns:
        """Create a column-oriented copy of a sequence of events."""
        column_values: Dict[str, List[Any]] = {}
        index = 0
        for event in events:
            for key, value in event.items():
                # Get the column to append 'value' to
                key_column = column_values.get(key)
                if key_column is None:
                    if columns is not None and key not in columns:
                        continue  # An excluded column
                    key_column = []
                    column_values[key] = key_column

                # Pad to the current index
                if len(key_column) < index:
                    key_column += [None] * (index - len(key_column))

                key_column.append(value)
            index += 1

        # Make sure all the columns finish the same length
        for key, key_column in column_values.items():
            if len(key_column) < index:
                key_column += [None] * (index - len(key_column))

        return cls(
            {key: list_to_array(value) for key, value in column_values.items()},
            column_order=None if columns is None else tuple(columns),
        )


class Log:
    """A friendly API for manipulating a single log file.

    The analysis support here is quite basic, focussed around handling of ordered
    heterogenous events. We suggest performing further analysis and plotting using
    {pandas, numpy, scipy, matplotlib, seaborn, etc.}
    """

    def __init__(self, events: Iterable[Event]):
        self.events = events
        self.header = get_header(events)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({_events_repr(self.events)})"

    def __getitem__(self, kind: str) -> Log:
        """Select events of a given kind from the log.

        Equivalent to `log.filter(kind)`.
        """
        if isinstance(kind, int):
            raise TypeError(
                "Cannot use a `Log` as an iterable - consider `Log.events` instead"
            )
        return self.filter(kind)

    @property
    def kinds(self) -> Set[Optional[str]]:
        """The set of {"kind": kind} from all events in the log."""
        return {typing.cast(Optional[str], event.get("kind")) for event in self.events}

    def cache(self) -> Log:
        """Create a log that is loaded into memory, for efficient multiple-traversal.

        Note that this does not change `self`, but returns a new cached Log.
        """
        if isinstance(self.events, (tuple, list)):
            return self
        return type(self)(tuple(self.events))

    def apply(self, *operations: ops.Operation) -> Log:
        """Create a transformed log view of this log.

        Note that the transformation will be executed whenever the log `events`
        are traversed.

        For example:

            log.apply(ops.count_if("step"))
        """
        return type(self)(Transform(self.events, ops.group(*operations)))

    def filter(self, predicate: ops.AutoPredicate) -> Log:
        """Create a filtered log view of this log."""
        return type(self)(Transform(self.events, ops.filter(predicate)))

    def to_columns(self, *columns: str) -> Columns:
        """Convert the log to a set of columns.

        If columns are specified, limit the output to these columns, with a
        defined order. Otherwise, the columns are autodetected from the log,
        and in unknown order.

        For example:

            step, loss = log["valid"].to_columns("step", "loss")
        """
        return Columns.from_events(self.events, columns if columns else None)

    def to_pandas(self, *columns: str) -> Any:
        """Convert the log to a pandas DataFrame.

        It's normally easiest to do this for a single event kind at a time.
        For example:

            dfv = log["valid"].to_pandas()
            dfs = log["step"].to_pandas()
        """
        return self.to_columns(*columns).to_pandas()


class LogSet:
    """A friendly "batched" API for manipulating a set of log files."""

    def __init__(self, logs: Sequence[Log]):
        self.logs = logs

    def __repr__(self) -> str:
        return f"{type(self).__name__}([{len(self.logs)}])"

    def __getitem__(self, kind_or_index: Union[str, int]) -> Union[LogSet, Log]:
        """Either filter log events (str) or index a single log (int)."""
        if isinstance(kind_or_index, str):
            return self.filter(kind_or_index)
        return self.logs[kind_or_index]

    def __len__(self) -> int:
        return len(self.logs)

    def __iter__(self) -> Iterator[Log]:
        return iter(self.logs)

    @property
    def events(self) -> Iterator[Event]:
        """A concatenated stream of all events from all logs."""
        for log in self.logs:
            yield from log.events

    @property
    def kinds(self) -> Set[Optional[str]]:
        """The set of all {"kind": kind} from all events in all logs."""
        return {kind for log in self.logs for kind in log.kinds}

    def cache(self) -> LogSet:
        """Create a log set that is loaded into memory, for efficient multiple-traversal.

        Note that this does not change `self`, but returns a new cached LogSet.
        """
        return type(self)(tuple(log.cache() for log in self.logs))

    def apply(self, *operations: ops.Operation) -> LogSet:
        """Create a (per-event) transformed view of this set of logs.

        For example:

            logs.apply(ops.header("id"), ops.count_if("step"))
        """
        return type(self)(tuple(log.apply(*operations) for log in self.logs))

    def filter(self, predicate: ops.AutoPredicate) -> LogSet:
        """Create a (per-event) filtered view of this set of logs.

        For example:

            logs.filter("valid")
        """
        return type(self)(tuple(log.filter(predicate) for log in self.logs))

    def to_columns(self, *columns: str) -> Columns:
        """Convert the logs to a set of columns.

        If columns are specified, limit the output to these columns, with a
        defined order. Otherwise, the columns are autodetected from the log,
        and in unknown order.
        """
        return Columns.from_events(self.events, columns if columns else None)

    def to_pandas(self, *columns: str) -> Any:
        """Convert the logs to a single pandas DataFrame.

        It's normally easiest to do this for a single event kind at a time.
        For example:

            logs["valid"].to_pandas()
        """
        return self.to_columns(*columns).to_pandas()


def open(  # pylint:disable=redefined-builtin
    path: str, load_args: Optional[Dict[str, Any]] = None
) -> Log:
    """Load a single Log from a local JSONLines file (e.g. written by logger.Log).

    For example:

        log = open("results/log.jsonl.gz")
    """
    return Log(JsonLinesFile(path, load_args=load_args))


def glob(pattern: str, recursive: bool = False) -> LogSet:
    """Load all logs matched by a local filesystem glob.

    For example:

        logs = glob("results/**/*.jsonl*", recursive=True)
    """
    return LogSet(tuple(open(f) for f in pyglob.glob(pattern, recursive=recursive)))



"""Friendly API for writing log files."""

from __future__ import annotations

import builtins
import contextlib
import datetime
import random
import time
from types import TracebackType
from typing import (
    Any,
    Callable,
    ContextManager,
    Dict,
    Iterable,
    Iterator,
    Optional,
    Protocol,
    Tuple,
    Type,
)

from . import io

Event = Dict[str, Any]
EventAnnotation = Callable[[Event], Optional[Event]]
Annotation = Callable[[Optional[Event]], Optional[EventAnnotation]]
LogLineScope = Callable[[Event], ContextManager[None]]


class Writer(Protocol):
    """Stream capable of writing general objects (e.g. io.JsonLinesIO)."""

    def write(self, obj: Event) -> None:
        """Write out a single event."""
        raise NotImplementedError

    def flush(self) -> None:
        """Flush the underlying stream - write out the buffered data."""
        raise NotImplementedError

    def close(self) -> None:
        """Close the underlying stream."""
        raise NotImplementedError


@contextlib.contextmanager
def add_duration(event: Event, ndigits: int = 3) -> Iterator[None]:
    """[LogLineScope] Add the time elapsed during a `log.adding` scope."""
    start = time.time()
    yield
    event["duration"] = round(time.time() - start, ndigits)


def set_time(header: Optional[Event], ndigits: int = 3) -> EventAnnotation:
    """[Annotation] Set the header date & elapsed fields."""
    start = time.time()
    if header is not None:
        header["time"] = datetime.datetime.now().isoformat()
    return lambda event: dict(elapsed=round(time.time() - start, ndigits))


def set_id(header: Optional[Event]) -> None:
    """[Annotation] Set the ID field of the header."""
    if header is not None:
        header["id"] = random.Random().randint(0, 1 << 64)


KIND_HEADER = "header"
DEFAULT_ANNOTATE: Tuple[Annotation, ...] = (set_time, set_id)
DEFAULT_SCOPES: Tuple[LogLineScope, ...] = (add_duration,)


class LogLine:
    """A mutable builder for a single line within a log.

    For example:

        with log.adding(kind="eval") as line:
            line.set(loss=compute_loss())
    """

    def __init__(self, log: Log, scopes: Iterable[LogLineScope], event: Event):
        self.log = log
        self.scopes = scopes
        self.event = event
        self.stack = contextlib.ExitStack()
        for scope in self.scopes:
            self.stack.enter_context(scope(self.event))
        self.finished = False

    def __enter__(self) -> LogLine:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.add_to_log()

    def set(self, **kwargs: Any) -> None:
        """Set/reset event parameters in the log"""
        if self.finished:
            raise ValueError("Log line set() after being added to the log")
        self.event.update(kwargs)

    def add_to_log(self) -> None:
        """Add the scope to the underlying log."""
        if self.finished:
            raise ValueError("Trying to add a log line to the log multiple times")
        self.stack.close()
        self.log.add(**self.event)
        self.finished = True


class Log:
    """A general-purpose event log.

    For example:

        with logger.open("log.jsonl") as log:
            log.add(kind="step", loss=step_loss())
            log.add(kind="step", loss=step_loss())
            with log.adding(kind="eval") as line:
                line.set(loss=eval_loss())
    """

    def __init__(  # pylint:disable=dangerous-default-value
        self,
        writer: Writer,
        header: Event = {},
        annotate: Iterable[Annotation] = (),
        default_annotate: bool = True,
    ):
        self.writer = writer
        header_event = None if header is None else dict(kind=KIND_HEADER, **header)
        all_annotate = (DEFAULT_ANNOTATE if default_annotate else ()) + tuple(annotate)
        self.annotators = [
            annotator
            for annotator in (fn(header_event) for fn in all_annotate)
            if annotator is not None
        ]
        if header_event is not None:
            self.add(**header_event)

    def __enter__(self) -> Log:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()

    def close(self) -> None:
        """Close the underlying stream (the log cannot be used afterwards).

        Note that this is called automatically if the log is used in a context manager
        (recommended) and it is an error to close() twice.
        """
        self.writer.close()

    def add(self, kind: Optional[str] = None, **args: Any) -> None:
        """Immediately add a single event to the log.

        The key "kind" is optional, but recommended if the log contains different kinds
        of events, to aid info extraction.

        Values must be serializable by the current writer (e.g. JSON-serializable).

        Note that extra keys may be added by "Annotations", which are set up in the
        Log constructor.
        """
        line = args if kind is None else dict(kind=kind, **args)
        for annotator in self.annotators:
            out = annotator(line)
            if out is not None:
                line.update(out)
        self.writer.write(line)
        self.writer.flush()

    def adding(
        self,
        kind: Optional[str] = None,
        _scopes: Iterable[LogLineScope] = (),
        _default_scopes: bool = True,
        **args: Any,
    ) -> LogLine:
        """Start building an event to add to the log.

        The key "kind" is optional, but recommended if the log contains different kinds
        of events, to aid info extraction.

        We recommend using this as a context manager:

            with log.adding(kind="eval") as line:
                line.set(loss=eval_loss())

        Scopes allow automated logging of before-and-after state, for example the
        (built-in, automatically added) scope `add_duration`. Disable built-in scopes
        with `default_scopes=False`.
        """
        all_scopes = tuple(_scopes) + (DEFAULT_SCOPES if _default_scopes else ())
        return LogLine(
            self,
            scopes=all_scopes,
            event=args if kind is None else dict(kind=kind, **args),
        )


class JsonLinesFileLog(Log):
    """A standard log that uses a `trainlog.io.JsonLinesIO` writer to a local file."""

    def __init__(
        self,
        path: str,
        gzip_on_close: bool = True,
        dump_args: Optional[Dict[str, Any]] = None,
        **args: Any,
    ):
        """Create a JsonLines writer the the given local path.

        If `gzip_on_close` is True, after the log is closed, the file is compressed
        to `<path>.gz` and the original deleted.
        """
        super().__init__(
            io.JsonLinesIO[Event](builtins.open(path, "w"), dump_args=dump_args),
            **args,
        )
        self.path = path
        self.gzip_on_close = gzip_on_close

    def close(self) -> None:
        """Close the underlying stream, optionally converting to gzip."""
        super().close()
        if self.gzip_on_close:
            io.gzip(self.path)


def open(  # pylint:disable=redefined-builtin
    path: str,
    _gzip_on_close: bool = True,
    _annotate: Iterable[Annotation] = (),
    _default_annotate: bool = True,
    _add_header: bool = True,
    **header: Any,
) -> JsonLinesFileLog:
    """Open a logger writing to a local JsonLines file.

    This is just syntactic sugar; if customization is needed, directly using
    `JsonLinesFileLog` or `Log` may be preferable.
    """
    if header and not _add_header:
        raise ValueError(
            "Trying to add keys to a header with _add_header=False.\n"
            f"Either remove the header {header.keys()} or use open(_add_header=True)."
        )
    return JsonLinesFileLog(
        path,
        gzip_on_close=_gzip_on_close,
        annotate=_annotate,
        default_annotate=_default_annotate,
        header=header if _add_header else None,
    )



"""Package setup."""

import pathlib

import setuptools  # type: ignore

import trainlog._version

project_dir = pathlib.Path(__file__).parent

setuptools.setup(
    name="trainlog",
    version=trainlog._version.__version__,  # pylint:disable=protected-access
    description="A simple logging library, designed for deep learning",
    long_description=(project_dir / "README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/DouglasOrr/TrainLog",
    author="Douglas Orr",
    license="MIT",
    packages=["trainlog"],
    install_requires=[
        req for req in (project_dir / "requirements.txt").read_text().split("\n") if req
    ],
)
